/*
 * Copyright (C) 2018 Magnus Bergman <magnus.bergman@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <stdint.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>                     /* strcasecmp() */

#include "debug.h"


typedef struct{
    uint8_t bitmap[8000];
    uint8_t screen[1000];
    uint8_t color[1000];
    uint8_t background;
}vmem_t;

struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    cairo_surface_t *surface;
};


static const char **
koala_get_mime_types()
{
    static const char *mime_types[] = {
        "image/x-koala",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
koala_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h;
    if(!strcasecmp(mime_type,"image/x-koala")){
        /* OK */
    }else{
        return NULL;
    }
    h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->surface = NULL;
    info->threadsafe = 1;
    return h;
}


static void
koala_free(abydos_plugin_handle_t *h)
{
    if(h->surface)
        cairo_surface_destroy(h->surface);
    free(h);
}


static uint32_t
get_color(vmem_t *vmem,int x,int y,int c)
{
    static const uint32_t color[16] = {
        0x000000, /* black */
        0xffffff, /* white */
        0x880000, /* dark red */
        0xaaffee, /* cyan */
        0xcc44cc, /* purple */
        0x00cc55, /* dark green */
        0x0000aa, /* dark blue */
        0xeeee77, /* yellow */
        0xdd8855, /* orange */
        0x664400, /* brown */
        0xff7777, /* light red */
        0x333333, /* dark gray */
        0x777777, /* mid gray */
        0xaaff66, /* light green */
        0x0088ff, /* light blue */
        0xbbbbbb  /* light gray */
    };
    switch(c){
    default:
    case 0:return color[vmem->background];
    case 1:return color[vmem->screen[x+40*y]>>4];
    case 2:return color[vmem->screen[x+40*y]&15];
    case 3:return color[vmem->color[x+40*y]&15];
    }
}


static cairo_surface_t *
c64_decode_image(abydos_plugin_handle_t *h,vmem_t *vmem)
{
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,160,200);
    uint8_t *src;
    uint32_t *dst;
    int stride;
    int x;
    int y;
    int i;
    src = vmem->bitmap;
    dst = (uint32_t*)cairo_image_surface_get_data(h->surface);
    stride = cairo_image_surface_get_stride(h->surface)/sizeof(uint32_t);
    for(y=0;y<25;++y){
        for(x=0;x<40;++x){
            for(i=0;i<8;++i){
                int pix = *src++;
                dst[x*4+0+(y*8+i)*stride] = get_color(vmem,x,y,(pix>>6) & 3);
                dst[x*4+1+(y*8+i)*stride] = get_color(vmem,x,y,(pix>>4) & 3);
                dst[x*4+2+(y*8+i)*stride] = get_color(vmem,x,y,(pix>>2) & 3);
                dst[x*4+3+(y*8+i)*stride] = get_color(vmem,x,y,(pix>>0) & 3);
            }
        }
    }
    return surface;
}


static int
koala_create_from_data(abydos_plugin_handle_t *h,const char *data,size_t len)
{
    DEBUG_p("begin (size = %d)",(int)len);
    if(len != 10003)
        return -1;
    h->info->width = 160;
    h->info->height = 200;
    h->info->pixel_ratio = 5.0/3;
    h->surface = c64_decode_image(h, (vmem_t*)&data[2]);
    return 0;
}


static cairo_surface_t *
koala_get_image_surface(abydos_plugin_handle_t *h,int page,cairo_bool_t *layer_visibility,int variant,double time,int frame)
{
    return cairo_surface_reference(h->surface);
}


const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = koala_get_mime_types,
        .new = koala_new,
        .free = koala_free,
        .create_from_data = koala_create_from_data,
        .get_image_surface = koala_get_image_surface
    };
    DEBUG_p("begin");
    return &p;
}
