/*
 * Copyright (C) 1999 Red Hat, Inc.
 * Copyright (C) 2018 Magnus Bergman <magnus.bergman@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
/*
 * Based on io-pnm.c from gdk-pixbuf (https://github.com/GNOME/gdk-pixbuf).
 *
 * Authors: Jeffrey Stedfast <fejj@helixcode.com>
 *          Michael Fulbright <drmike@redhat.com>
 */
#include <abydos-plugin.h>
#include <assert.h>
#include <ctype.h>                      /* isspace() */
#include <fcntl.h>                      /* O_NONBLOCK */
#include <limits.h>                     /* LONG_MAX */
#include <netinet/in.h>                 /* ntohs() */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/signal.h>
#include <sys/wait.h>
#include <unistd.h>                     /* dup2() */

#include "buffer.h"
#include "debug.h"


#define ATARI_PIXEL_RATIO (5.0/6)
//#define BUFFER_SIZE 4096


#ifndef MIN
# define MIN(a,b) ((a)<(b)?(a):(b))
#endif


typedef enum {
    PNM_OK = 0,
    PNM_CONTINUE,
    PNM_ERROR
} result_t;

typedef enum {
    PNM_FORMAT_PGM = 1,
    PNM_FORMAT_PGM_RAW,
    PNM_FORMAT_PPM,
    PNM_FORMAT_PPM_RAW,
    PNM_FORMAT_PBM,
    PNM_FORMAT_PBM_RAW
} pnm_format_t;

typedef void(*draw_func_t)(abydos_plugin_handle_t *,int pixels,int bytes);

struct _abydos_plugin_handle_t{
    cairo_surface_t *pixbuf;
    uint8_t *pixels;        /* incoming pixel data buffer */
    int rowstride;
    abydos_plugin_info_t *info;

    pnm_format_t type;
    draw_func_t draw_func;
    int maxval;

    /* external converter */
    const char *converter;
    int input[2];
    int output[2];
    int pid;

    /* progressive loading */
    abydos_buffer_t inbuf;
    abydos_plugin_info_func_t info_func;
    abydos_plugin_progress_func_t progress_func;
    void *userdata;
    enum{
        S_HEADER,
        S_PRESCAN,
        S_BODY,
        S_DONE
    }state;
    //bool got_header;   /* have we loaded pnm header? */
    unsigned int scan_state;
    //uint32_t *dptr;          /* current position in pixbuf */
    int output_row;      /* last row to be completed */
    int output_col;
    //bool did_prescan;  /* are we in image data yet? */

    //GError **error;

};


/* skip over whitespace and comments in input buffer */
static result_t
pnm_skip_whitespace (abydos_buffer_t *inbuf)
{
    enum {S_NONE,S_COMMENT} state = S_NONE;
    uint8_t ch;

    while (abydos_buffer_size(inbuf) > 0) {
        ch = *abydos_buffer_get(inbuf,1);
        switch(state){
        case S_NONE:
            if('#'==ch){
                state = S_COMMENT;
            }else if(!isspace(ch)){
                abydos_buffer_add_back(inbuf,&ch,1);
                return PNM_OK;
            }
            break;
        case S_COMMENT:
            if('\n'==ch){
                state = S_NONE;
            }else if(abydos_buffer_size(inbuf) <= 0){
                /* couldn't read whole comment */
                abydos_buffer_add_back(inbuf,(const uint8_t*)"#",1);
            }
            break;
        }
    }
    return PNM_CONTINUE;
}


/* read next number from buffer */
static result_t
pnm_read_next_value (abydos_buffer_t *inbuf, int max_length, int *r_value)
{
    int value = 0;
    result_t res;
    int i;

    assert(r_value);

    /* skip white space */
    res = pnm_skip_whitespace (inbuf);
    if (res != PNM_OK)
        return res;

    for(i=0;max_length<0 || i<max_length;++i){
        int ch = abydos_buffer_peek_at(inbuf,i);
        if(ch>='0' && ch<='9'){
            value *= 10;
            value += ch-'0';
        }else if(isspace(ch) || '#'==ch){
            abydos_buffer_skip(inbuf,i);
            *r_value = value;
            return PNM_OK;
        }else if(ch < 0){
            return PNM_CONTINUE;
        }else{
            return PNM_ERROR;
        }
    }
    return PNM_OK;
}


/* returns PNM_OK, PNM_SUSPEND, or PNM_FATAL_ERR */
static result_t
pnm_read_header (abydos_plugin_handle_t *h)
{
    abydos_buffer_t *inbuf = &h->inbuf;
    result_t res;

    if (!h->type) {
        const uint8_t *src = abydos_buffer_get(&h->inbuf,2);
        if(!src)
            return PNM_CONTINUE;
        /* file must start with a 'P' followed by a numeral  */
        /* so loop till we get enough data to determine type */

        if ('P' != src[0]) {
            /* _("PNM file has an incorrect initial byte")); */
            return PNM_ERROR;
        }

        switch (src[1]) {
        case '1':
            h->type = PNM_FORMAT_PBM;
            break;
        case '2':
            h->type = PNM_FORMAT_PGM;
            break;
        case '3':
            h->type = PNM_FORMAT_PPM;
            break;
        case '4':
            h->type = PNM_FORMAT_PBM_RAW;
            break;
        case '5':
            h->type = PNM_FORMAT_PGM_RAW;
            break;
        case '6':
            h->type = PNM_FORMAT_PPM_RAW;
            break;
        default:
            /* _("PNM file is not in a recognized PNM subformat")); */
            return PNM_ERROR;
        }
    }

    if (!h->info->width) {
        /* read the pixmap width */
        int width = 0;

        res = pnm_read_next_value (inbuf, -1, &width);
        if (res != PNM_OK)
            return res;

        if (width<=0) {
            /* _("PNM file has an image width of 0")); */
            return PNM_ERROR;
        }
        DEBUG_p("got width: %d",width);
        h->info->width = width;
    }

    if (!h->info->height) {
        /* read the pixmap height */
        int height = 0;

        res = pnm_read_next_value (inbuf, -1, &height);
        if (res != PNM_OK)
            return res;

        if (!height) {
            /* _("PNM file has an image height of 0")); */
            return PNM_ERROR;
        }
        DEBUG_p("got height: %d",height);
        h->info->height = height;
    }

    switch (h->type) {
    case PNM_FORMAT_PPM:
    case PNM_FORMAT_PPM_RAW:
    case PNM_FORMAT_PGM:
    case PNM_FORMAT_PGM_RAW:
        if (!h->maxval) {
            res = pnm_read_next_value(inbuf, -1, &h->maxval);
            if (res != PNM_OK)
                return res;

            if (h->maxval <= 0) {
                /* _("Maximum color value in PNM file is 0")); */
                return PNM_ERROR;
            }

            if (h->maxval > 65535) {
                /* _("Maximum color value in PNM file is too large") */
                return PNM_ERROR;
            }

        }
        break;
    default:
        break;
    }

    return PNM_OK;
}


static void
draw_pbm_raw(abydos_plugin_handle_t *h,int pixels,int bytes){
    const uint8_t *src = abydos_buffer_get(&h->inbuf,bytes);
    uint32_t *dst = (uint32_t*)(h->pixels + h->output_row * h->rowstride);
    int i;
    dst += h->output_col;
    for(i=0; i < pixels ; ++i){
        int bits = *src++;
        int j;
        for(j=0; i < 8; ++j){
            bits <<= 1;
            *dst++ = bits&0x100 ? 0 : 0xffffff;
        }
    }
}

static void
draw_pgm_raw_8(abydos_plugin_handle_t *h,int pixels,int bytes){
    const uint8_t *src = abydos_buffer_get(&h->inbuf,bytes);
    uint32_t *dst = (uint32_t*)(h->pixels + h->output_row * h->rowstride);
    int i;
    dst += h->output_col;
    for(i=0; i< pixels ; ++i){
        int v = *src++;
        *dst++ = (v << 16) | (v << 8) | v;
    }
}

static void
draw_pgm_raw_8_scale(abydos_plugin_handle_t *h,int pixels,int bytes){
    const uint8_t *src = abydos_buffer_get(&h->inbuf,bytes);
    uint32_t *dst = (uint32_t*)(h->pixels + h->output_row * h->rowstride);
    int i;
    dst += h->output_col;
    for (i = 0; i < pixels; ++i) {
        int v = 255 * (*src++) / h->maxval;
        *dst++ = (v << 16) | (v << 8) | v;
    }
}

static void
draw_pgm_raw_16(abydos_plugin_handle_t *h,int pixels,int bytes){
    const uint16_t *src = (const uint16_t*)abydos_buffer_get(&h->inbuf,bytes);
    uint32_t *dst = (uint32_t*)(h->pixels + h->output_row * h->rowstride);
    int i;
    dst += h->output_col;
    for(i=0; i < pixels ; ++i){
        int v = ntohs(*src++) >> 8;
        *dst++ = (v << 16) | (v << 8) | v;
    }
}

static void
draw_pgm_raw_16_scale(abydos_plugin_handle_t *h,int pixels,int bytes){
    const uint16_t *src = (const uint16_t*)abydos_buffer_get(&h->inbuf,bytes);
    uint32_t *dst = (uint32_t*)(h->pixels + h->output_row * h->rowstride);
    int i;
    dst += h->output_col;
    for(i=0; i < pixels ; ++i){
        int v = 255 * ntohs(*src++) / h->maxval;
        *dst++ = (v << 16) | (v << 8) | v;
    }
}

static void
draw_ppm_raw_8(abydos_plugin_handle_t *h,int pixels,int bytes){
    const uint8_t *src = abydos_buffer_get(&h->inbuf,bytes);
    uint32_t *dst = (uint32_t*)(h->pixels + h->output_row * h->rowstride);
    int i;
    dst += h->output_col;
    for(i=0; i< pixels ; ++i){
        int r = *src++;
        int g = *src++;
        int b = *src++;
        *dst++ = (r << 16) | (g << 8) | b;
    }
}

static void
draw_ppm_raw_8_scale(abydos_plugin_handle_t *h,int pixels,int bytes){
    const uint8_t *src = abydos_buffer_get(&h->inbuf,bytes);
    uint32_t *dst = (uint32_t*)(h->pixels + h->output_row * h->rowstride);
    int i;
    dst += h->output_col;
    for (i = 0; i < pixels; ++i) {
        int r = 255 * (*src++) / h->maxval;
        int g = 255 * (*src++) / h->maxval;
        int b = 255 * (*src++) / h->maxval;
        *dst++ = (r << 16) | (g << 8) | b;
    }
}

static void
draw_ppm_raw_16(abydos_plugin_handle_t *h,int pixels,int bytes){
    const uint16_t *src = (const uint16_t*)abydos_buffer_get(&h->inbuf,bytes);
    uint32_t *dst = (uint32_t*)(h->pixels + h->output_row * h->rowstride);
    int i;
    dst += h->output_col;
    for(i=0; i < pixels ; ++i){
        int r = ntohs(*src++) >> 8;
        int g = ntohs(*src++) >> 8;
        int b = ntohs(*src++) >> 8;
        *dst++ = (r << 16) | (g << 8) | b;
    }
}

static void
draw_ppm_raw_16_scale(abydos_plugin_handle_t *h,int pixels,int bytes){
    const uint16_t *src = (const uint16_t*)abydos_buffer_get(&h->inbuf,bytes);
    uint32_t *dst = (uint32_t*)(h->pixels + h->output_row * h->rowstride);
    int i;
    dst += h->output_col;
    for(i=0; i < pixels ; ++i){
        int r = 255 * ntohs(*src++) / h->maxval;
        int g = 255 * ntohs(*src++) / h->maxval;
        int b = 255 * ntohs(*src++) / h->maxval;
        *dst++ = (r << 16) | (g << 8) | b;
    }
}


static result_t
pnm_read_raw_scanline (abydos_plugin_handle_t *h)
{
    int pixels;
    int bytes;;

    switch (h->type) {
    case PNM_FORMAT_PBM_RAW:
        pixels = abydos_buffer_size(&h->inbuf) * 8;
        break;
    case PNM_FORMAT_PGM_RAW:
        pixels = abydos_buffer_size(&h->inbuf);
        break;
    case PNM_FORMAT_PPM_RAW:
        pixels = abydos_buffer_size(&h->inbuf) / 3;
        break;
    default:
        /* _("Raw PNM image type is invalid") */
        return PNM_ERROR;
    }
    if(h->maxval>255)
        pixels/=2;

    pixels = MIN (pixels, h->info->width - h->output_col);

    if (pixels <= 0)
        return PNM_CONTINUE;

    switch (h->type) {
    case PNM_FORMAT_PBM_RAW:
        bytes = (pixels-1) / 8 + 1;
        break;
    case PNM_FORMAT_PGM_RAW:
        bytes = pixels;
        break;
    case PNM_FORMAT_PPM_RAW:
        bytes = pixels * 3;
        break;
    default:
        /* _("Raw PNM image type is invalid") */
        return PNM_ERROR;
    }
    if(h->maxval>255)
        bytes*=2;

    h->draw_func(h,pixels,bytes);

    h->output_col += pixels;
    if (h->output_col >= h->info->width) {
        h->output_col = 0;
        h->output_row++;
    } else {
        return PNM_CONTINUE;
    }

    return PNM_OK;
}


static result_t
pnm_read_ascii_scanline (abydos_plugin_handle_t *h)
{
    abydos_buffer_t *inbuf = &h->inbuf;
    int value, numval, i;
    uint8_t data;
    uint8_t mask;
    uint32_t *dptr;
    result_t res;
    int max_length;

    data = mask = 0;

    //h->dptr = h->pixels + h->output_row * h->rowstride;

    switch (h->type) {
    case PNM_FORMAT_PBM:
        max_length = 1;
        numval = MIN (8, h->info->width - h->output_col);
        break;
    case PNM_FORMAT_PGM:
        max_length = -1;
        numval = 1;
        break;
    case PNM_FORMAT_PPM:
        max_length = -1;
        numval = 3;
        break;

    default:
        /* _("PNM image format is invalid") */
        return PNM_ERROR;
    }

    //dptr = h->dptr + offset + h->scan_state;

    for (;;) {
        if (h->type == PNM_FORMAT_PBM) {
            mask = 0x80;
            data = 0;
            numval = MIN (8, h->info->width - h->output_col);
        }

        for (i = h->scan_state; i < numval; i++) {
            res = pnm_read_next_value (inbuf, max_length, &value);
            if (res != PNM_OK) {
                /* save state and return */
                h->scan_state = i;
                return res;
            }

            switch (h->type) {
            case PNM_FORMAT_PBM:
                if (value)
                    data |= mask;
                mask >>= 1;

                break;
            case PNM_FORMAT_PGM:
            case PNM_FORMAT_PPM:
                /* scale the color up or down to an 8-bit color depth */
                if (value > h->maxval)
                    *dptr++ = 255;
                else
                    *dptr++ = 255 * value / h->maxval;
                break;
            default:
                /* _("PNM image format is invalid") */
                return PNM_ERROR;
                break;
            }
        }

        h->scan_state = 0;

        if (h->type == PNM_FORMAT_PBM) {
            *dptr++ = data;
            h->output_col += numval;
        } else {
            h->output_col++;
        }

        if (h->output_col >= h->info->width) {
#if 0
            if (h->type == PNM_FORMAT_PBM)
                explode_bitmap_into_buf (h);
            else if (h->type == PNM_FORMAT_PGM)
                explode_gray_into_buf (h);
#endif

            h->output_col = 0;
            h->output_row++;
            break;
        }
    }

    return PNM_OK;
}


static result_t
pnm_read_scanline (abydos_plugin_handle_t *h)
{
    result_t res;

    /* read in image data */
    /* for raw formats this is trivial */
    switch (h->type) {
    case PNM_FORMAT_PBM_RAW:
    case PNM_FORMAT_PGM_RAW:
    case PNM_FORMAT_PPM_RAW:
        res = pnm_read_raw_scanline (h);
        if (res != PNM_OK)
            return res;
        break;
    case PNM_FORMAT_PBM:
    case PNM_FORMAT_PGM:
    case PNM_FORMAT_PPM:
        res = pnm_read_ascii_scanline (h);
        if (res != PNM_OK)
            return res;
        break;
    default:
        /* _("PNM image loader does not support this PNM subformat")); */
        return PNM_ERROR;
    }

    return PNM_OK;
}


/*
 * context - from image_begin_load
 * buf - new image data
 * size - length of new image data
 *
 * append image data onto inrecrementally built output image
 */
static result_t
pnm_feed (abydos_plugin_handle_t *h,const char *data,size_t len)
{
    abydos_buffer_t *inbuf = &h->inbuf;
    result_t res;

    abydos_buffer_add(inbuf,(const uint8_t*)data,len);
    while(abydos_buffer_size(inbuf) > 0) {
        switch(h->state){
        case S_HEADER:
            /* get header if needed */
            res = pnm_read_header (h);
            if (res != PNM_OK)
                goto done;
            if(h->info_func)
                h->info_func(h->userdata);
            switch(h->type){
            case PNM_FORMAT_PBM:
            case PNM_FORMAT_PBM_RAW:
                h->draw_func = draw_pbm_raw;
                break;
            case PNM_FORMAT_PGM:
            case PNM_FORMAT_PGM_RAW:
                if(h->maxval > 255){
                    if(h->maxval == 65535) {
                        h->draw_func = draw_pgm_raw_16;
                    }else{
                        h->draw_func = draw_pgm_raw_16_scale;
                    }
                }else{
                    if (h->maxval == 255) {
                        h->draw_func = draw_pgm_raw_8;
                    } else {
                        h->draw_func = draw_pgm_raw_8_scale;
                    }
                }
                break;
            case PNM_FORMAT_PPM:
            case PNM_FORMAT_PPM_RAW:
                if(h->maxval > 255){
                    if(h->maxval == 65535) {
                        h->draw_func = draw_ppm_raw_16;
                    }else{
                        h->draw_func = draw_ppm_raw_16_scale;
                    }
                }else{
                    if (h->maxval == 255) {
                        h->draw_func = draw_ppm_raw_8;
                    } else {
                        h->draw_func = draw_ppm_raw_8_scale;
                    }
                }
                break;
            }
            h->state = S_PRESCAN;
            /* break; */
        case S_PRESCAN:
            /* scan until we hit image data */
            switch (h->type) {
            case PNM_FORMAT_PBM_RAW:
            case PNM_FORMAT_PGM_RAW:
            case PNM_FORMAT_PPM_RAW:
                if (abydos_buffer_size(inbuf) <= 0){
                    res = PNM_CONTINUE;
                    goto done;
                }
                /* raw formats require exactly one whitespace */
                if (!isspace(*abydos_buffer_get(inbuf,1))){
                    /* _("Raw PNM formats require exactly one whitespace before sample data")); */
                    return PNM_ERROR;
                }
                break;
            default:
                res = pnm_skip_whitespace(inbuf);
                if(res != PNM_OK)
                    goto done;
                break;
            }
            h->output_row = 0;
            h->output_col = 0;

            h->pixbuf = cairo_image_surface_create(CAIRO_FORMAT_RGB24,h->info->width,h->info->height);
            if (!h->pixbuf) {
                /* _("Insufficient memory to load PNM file") */
                return PNM_ERROR;
            }
            h->pixels = cairo_image_surface_get_data(h->pixbuf);
            h->rowstride = cairo_image_surface_get_stride(h->pixbuf);

            h->state = S_BODY;
            /* break; */
        case S_BODY:
            /* if we got here we're reading image data */
            while (h->output_row < h->info->height) {
                res = pnm_read_scanline (h);
                if (res != PNM_OK)
                    goto done;
                if (h->progress_func) {
                    cairo_rectangle_int_t rect;
                    rect.x = 0;
                    rect.y = h->output_row-1;
                    rect.width = h->info->width;
                    rect.height = 1;
                    h->progress_func (h->userdata,&rect);
                }
            }
            h->state = S_DONE;
            /* break */
        case S_DONE:
            /* the image is fully loaded */
            res = PNM_OK;
            goto done;
        }
    }
done:
    abydos_buffer_privatize(inbuf);
    return res;
}



static result_t
read_and_feed(abydos_plugin_handle_t *h)
{
    for(;;){
        char buffer[ABYDOS_BUFFER_SIZE];
        ssize_t bytes = read(h->input[0],buffer,ABYDOS_BUFFER_SIZE);
        result_t res;
        if(bytes<=0)
            return PNM_CONTINUE;
        res = pnm_feed(h,buffer,bytes);
        if(res != PNM_CONTINUE)
            return res;
    }
}



static const char **
netpbm_get_mime_types()
{
    static const char *mime_types[] = {
        "image/vnd.zbrush.pcx",
        "image/x-ilbm",
        "image/x-neo",
        "image/x-neochrome",
        "image/x-pc1",
        "image/x-pc2",
        "image/x-pc3",
        "image/x-pcx",
        //"image/x-pi1",
        //"image/x-pi2",
        //"image/x-pi3",
        "image/x-portable-anymap",
        "image/x-portable-bitmap",
        "image/x-portable-graymap",
        "image/x-portable-pixmap",
        "image/x-spc",
        "image/x-spu",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
netpbm_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->type = 0;
    h->info->pixel_ratio = 1;
    if(!strcasecmp(mime_type,"image/vnd.zbrush.pcx")
      |!strcasecmp(mime_type,"image/x-pcx")){
        h->converter = "pcxtoppm";
    }else if(!strcasecmp(mime_type,"image/x-ilbm")){
        h->converter = "ilbmtoppm";
    }else if(!strcasecmp(mime_type,"image/x-neo")
            |!strcasecmp(mime_type,"image/x-neochrome")){
        h->converter = "neotoppm";
        h->info->pixel_ratio = ATARI_PIXEL_RATIO;
    }else if(!strcasecmp(mime_type,"image/x-pc1")){
        h->converter = "pc1toppm";
        h->info->pixel_ratio = ATARI_PIXEL_RATIO;
    }else if(!strcasecmp(mime_type,"image/x-pi1")){
        h->converter = "pi1toppm";
        h->info->pixel_ratio = ATARI_PIXEL_RATIO;
    }else if(!strcasecmp(mime_type,"image/x-portable-anymap")
            |!strcasecmp(mime_type,"image/x-portable-bitmap")
            |!strcasecmp(mime_type,"image/x-portable-graymap")
            |!strcasecmp(mime_type,"image/x-portable-pixmap")){
        h->converter = NULL;
    }else if(!strcasecmp(mime_type,"image/x-spc")){
        h->converter = "spctoppm";
        h->info->pixel_ratio = ATARI_PIXEL_RATIO;
    }else if(!strcasecmp(mime_type,"image/x-spu")){
        h->converter = "sputoppm";
        h->info->pixel_ratio = ATARI_PIXEL_RATIO;
    }else{
        free(h);
        return NULL;
    }
    info->threadsafe = 1;
    return h;
}


static void
netpbm_free(abydos_plugin_handle_t *h)
{
    if(h->pixbuf)
        cairo_surface_destroy(h->pixbuf);
    free(h);
}


static void
netpbm_progressive_begin(abydos_plugin_handle_t *h,abydos_plugin_info_func_t info_func,abydos_plugin_progress_func_t progress_func,void *userdata)
{
    h->info_func = info_func;
    h->progress_func = progress_func;
    h->userdata = userdata;
    h->maxval = 0;
    h->pixbuf = NULL;
    h->pixels = NULL;
    h->state = S_HEADER;
    //h->got_header = false;
    //h->did_prescan = false;
    h->scan_state = 0;

    abydos_buffer_clear(&h->inbuf);

    if(h->converter){
        int flags;
        pipe(h->input);
        pipe(h->output);
        h->pid = fork();
        if(0==h->pid){
            close(0);
            dup2(h->output[0],0);
            close(1);
            dup2(h->input[1],1);
            execlp(h->converter,h->converter,NULL);
        }
        close(h->output[0]);
        close(h->input[1]);
        flags = fcntl(h->input[0],F_GETFL,0);
        fcntl(h->input[0], F_SETFL, flags | O_NONBLOCK);
    }
}


static int
netpbm_progressive_feed(abydos_plugin_handle_t *h, const char *data, size_t len)
{
    DEBUG_p("writing %i bytes",(int)len);
    if(h->converter){
        write(h->output[1],data,len);
        read_and_feed(h);
    }else{
        pnm_feed(h,data,len);
    }
    return S_DONE == h->state;
}


static int
netpbm_progressive_end(abydos_plugin_handle_t *h)
{
    if(h->converter){
        int flags;
        DEBUG_p("closing pipe");
        close(h->output[1]);
        /* Make input pipe blocking again to let the converter have a
           chance to finish. */
        flags = fcntl(h->input[0],F_GETFL,0);
        fcntl(h->input[0], F_SETFL, flags & ~O_NONBLOCK);
        read_and_feed(h);
        kill(h->pid,SIGTERM);
        waitpid(h->pid,NULL,0);
        close(h->input[0]);
    }
#if 0
    /* We should ignore trailing newlines and we can't
       generally complain about trailing stuff at all, since
       pnm allows to put multiple images in a file
    */
    if (context->inbuf.nbytes > 0) {
         /* _("Unexpected end of PNM image data") */
        retval = false;
    }
#endif
    return S_DONE == h->state;
}


static cairo_surface_t *
netpbm_get_image_surface(abydos_plugin_handle_t *h,int page,cairo_bool_t *layer_visibility,int variant, double time, int frame)
{
    return cairo_surface_reference(h->pixbuf);
}


const abydos_plugin_interface_t *
Abydos0(){
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = netpbm_get_mime_types,
        .new = netpbm_new,
        .free = netpbm_free,
        .progressive_begin = netpbm_progressive_begin,
        .progressive_feed = netpbm_progressive_feed,
        .progressive_end = netpbm_progressive_end,
        .get_image_surface = netpbm_get_image_surface
    };
    DEBUG_p("begin");
    return &p;
}
