/*
 * Copyright (C) 2018 Magnus Bergman <magnus.bergman@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <librsvg/rsvg.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>                     /* strcasecmp() */


struct _abydos_plugin_handle_t{
    abydos_plugin_info_t *info;
    RsvgHandle *handle;
    abydos_plugin_info_func_t info_func;
    abydos_plugin_progress_func_t progress_func;
    void *userdata;
};


static const char **
svg_get_mime_types()
{
    static const char *mime_types[] = {
        "image/svg+xml",
        "application/svg+xml",
        NULL
    };
    return mime_types;
}


static abydos_plugin_handle_t *
svg_new(const char *mime_type, abydos_plugin_info_t *info)
{
    abydos_plugin_handle_t *h;
    if(!strcasecmp(mime_type,"image/svg+xml")
      |!strcasecmp(mime_type,"application/svg+xml")){
        /* OK */
    }else{
        return NULL;
    }
    h = (abydos_plugin_handle_t*)malloc(sizeof(abydos_plugin_handle_t));
    h->info = info;
    h->handle = NULL;
    info->scalable_size = 1;
    info->threadsafe = 1;
    return h;
}


static void
svg_free(abydos_plugin_handle_t *h)
{
    if(h->handle)
        g_object_unref(h->handle);
    free(h);
}



static void
svg_progressive_begin(abydos_plugin_handle_t *h,
                      abydos_plugin_info_func_t info_func,
                      abydos_plugin_progress_func_t progress_func,
                      void *userdata)
{
    h->handle = rsvg_handle_new_with_flags(RSVG_HANDLE_FLAG_KEEP_IMAGE_DATA);
    h->info_func = info_func;
    h->progress_func = progress_func;
    h->userdata = userdata;
}


static int
svg_progressive_feed(abydos_plugin_handle_t *h,const char* data, size_t len)
{
    /* FIXME!! set error from gerror */
    return rsvg_handle_write(h->handle,(const unsigned char*)data,len,NULL) ? 0 : -1;
}


static int
svg_progressive_end(abydos_plugin_handle_t *h)
{
    RsvgDimensionData dim;
    if(!rsvg_handle_close(h->handle,NULL)){
        /* FIXME!! set error from gerror */
        return -1;
    }
    rsvg_handle_get_dimensions(h->handle,&dim);
    h->info->width = dim.width;
    h->info->height = dim.height;
    if(h->info_func)
        h->info_func(h->userdata);
    if(h->progress_func){
        cairo_rectangle_int_t rect;
        rect.x = 0;
        rect.y = 0;
        rect.width = dim.width;
        rect.height = dim.height;
        h->progress_func(h->userdata,&rect);
    }
    return 0;
}



static void
svg_render(abydos_plugin_handle_t *h, cairo_t *cr,int page,cairo_bool_t *visible_layers,int prefered_variant,double time,int frame)
{
    rsvg_handle_render_cairo(h->handle, cr);
}



const abydos_plugin_interface_t *
Abydos0()
{
    static const abydos_plugin_interface_t p = {
        .version = ABYDOS_PLUGIN_INTERFACE_VERSION,
        .get_mime_types = svg_get_mime_types,
        .new = svg_new,
        .free = svg_free,
        .progressive_begin = svg_progressive_begin,
        .progressive_feed = svg_progressive_feed,
        .progressive_end = svg_progressive_end,
        .render = svg_render
    };
    return &p;
}
